import uuid
from datetime import datetime
from typing import Any
from typing import List
from typing import Optional
from typing import Union

from lib.core.enums import AnnotationStatus
from pydantic import BaseModel
from pydantic import Extra
from pydantic import Field
from pydantic import StrictBool
from pydantic import StrictFloat
from pydantic import StrictInt
from pydantic import StrictStr
from pydantic.datetime_parse import parse_datetime


class StringDate(datetime):
    @classmethod
    def __get_validators__(cls):
        yield parse_datetime
        yield cls.validate

    @classmethod
    def validate(cls, v: datetime):
        return v.isoformat()


class SubSetEntity(BaseModel):
    id: Optional[int]
    name: str

    class Config:
        extra = Extra.ignore


class TimedBaseModel(BaseModel):
    createdAt: StringDate = Field(None, alias="createdAt")
    updatedAt: StringDate = Field(None, alias="updatedAt")


class BaseItemEntity(TimedBaseModel):
    name: str
    path: Optional[str] = Field(
        None, description="Item’s path in SuperAnnotate project"
    )
    url: Optional[str] = Field(description="Publicly available HTTP address")
    annotator_email: Optional[str] = Field(description="Annotator email")
    qa_email: Optional[str] = Field(description="QA email")
    annotation_status: AnnotationStatus = Field(description="Item annotation status")
    entropy_value: Optional[float] = Field(description="Priority score of given item")
    createdAt: str = Field(description="Date of creation")
    updatedAt: str = Field(description="Update date")
    custom_metadata: Optional[dict]

    class Config:
        extra = Extra.allow

    def add_path(self, project_name: str, folder_name: str):
        self.path = (
            f"{project_name}{f'/{folder_name}' if folder_name != 'root' else ''}"
        )
        return self

    @staticmethod
    def map_fields(entity: dict) -> dict:
        entity["url"] = entity.get("path")
        entity["path"] = None
        entity["annotator_email"] = entity.get("annotator_id")
        entity["qa_email"] = entity.get("qa_id")
        return entity


class AttachmentEntity(BaseModel):
    name: Optional[str] = Field(default_factory=lambda: str(uuid.uuid4()))
    url: str

    class Config:
        extra = Extra.ignore


class SettingEntity(BaseModel):
    id: Optional[int]
    project_id: Optional[int]
    attribute: str
    value: Union[StrictStr, StrictInt, StrictFloat, StrictBool]

    class Config:
        extra = Extra.ignore

    def __copy__(self):
        return SettingEntity(attribute=self.attribute, value=self.value)


class ProjectEntity(TimedBaseModel):
    id: Optional[int]
    team_id: Optional[int]
    name: Optional[str]
    type: Optional[int]
    description: Optional[str]
    instructions_link: Optional[str]
    creator_id: Optional[str]
    entropy_status: Optional[int]
    sharing_status: Optional[int]
    status: Optional[int]
    folder_id: Optional[int]
    sync_status: Optional[int]
    upload_state: Optional[int]
    users: Optional[List[Any]] = []
    unverified_users: Optional[List[Any]] = []
    contributors: Optional[List[Any]] = []
    settings: Optional[List[SettingEntity]] = []
    classes: Optional[List[Any]] = []
    workflows: Optional[List[Any]] = []
    completed_images_count: Optional[int] = Field(None, alias="completedImagesCount")
    root_folder_completed_images_count: Optional[int] = Field(
        None, alias="rootFolderCompletedImagesCount"
    )

    class Config:
        extra = Extra.ignore

    def __copy__(self):
        return ProjectEntity(
            team_id=self.team_id,
            name=self.name,
            type=self.type,
            description=self.description,
            instructions_link=self.instructions_link
            if self.description
            else f"Copy of {self.name}.",
            status=self.status,
            folder_id=self.folder_id,
            users=self.users,
            upload_state=self.upload_state,
        )



import copy
from typing import List
from typing import Optional

import superannotate.lib.core as constants
from lib.core.conditions import Condition
from lib.core.conditions import CONDITION_EQ as EQ
from lib.core.entities import AttachmentEntity
from lib.core.entities import BaseItemEntity
from lib.core.entities import DocumentEntity
from lib.core.entities import FolderEntity
from lib.core.entities import ImageEntity
from lib.core.entities import ProjectEntity
from lib.core.entities import SubSetEntity
from lib.core.entities import TmpImageEntity
from lib.core.entities import VideoEntity
from lib.core.exceptions import AppException
from lib.core.exceptions import AppValidationException
from lib.core.exceptions import BackendError
from lib.core.reporter import Reporter
from lib.core.repositories import BaseReadOnlyRepository
from lib.core.response import Response
from lib.core.serviceproviders import SuperannotateServiceProvider
from lib.core.usecases.base import BaseReportableUseCase
from lib.core.usecases.base import BaseUseCase
from superannotate.logger import get_default_logger

logger = get_default_logger()


class GetBulkItems(BaseUseCase):
    def __init__(
        self,
        service: SuperannotateServiceProvider,
        project_id: int,
        team_id: int,
        folder_id: int,
        items: List[str],
    ):
        super().__init__()
        self._service = service
        self._project_id = project_id
        self._team_id = team_id
        self._folder_id = folder_id
        self._items = items
        self._chunk_size = 500

    def execute(self):
        res = []
        for i in range(0, len(self._items), self._chunk_size):
            response = self._service.get_bulk_items(
                project_id=self._project_id,
                team_id=self._team_id,
                folder_id=self._folder_id,
                items=self._items[i : i + self._chunk_size],  # noqa: E203
            )

            if not response.ok:
                raise AppException(response.error)
            # TODO stop using Image Entity when it gets deprecated and from_dict gets implemented for items
            res += [ImageEntity.from_dict(**item) for item in response.data]
        self._response.data = res
        return self._response


class GetItem(BaseReportableUseCase):
    def __init__(
        self,
        reporter: Reporter,
        project: ProjectEntity,
        folder: FolderEntity,
        backend_client: SuperannotateServiceProvider,
        item_name: str,
        include_custom_metadata: bool,
    ):
        super().__init__(reporter)
        self._project = project
        self._folder = folder
        self._backend_client = backend_client
        self._item_name = item_name
        self._include_custom_metadata = include_custom_metadata

    def validate_project_type(self):
        if (
            self._project.type == constants.ProjectType.PIXEL.value
            and self._include_custom_metadata
        ):
            raise AppException(constants.METADATA_DEPRICATED_FOR_PIXEL)

    @staticmethod
    def serialize_entity(entity: BaseItemEntity, project: ProjectEntity):
        entity = BaseItemEntity(**BaseItemEntity.map_fields(entity.dict()))
        if project.upload_state != constants.UploadState.EXTERNAL.value:
            entity.url = None
        if project.type in (
            constants.ProjectType.VECTOR.value,
            constants.ProjectType.PIXEL.value,
        ):
            tmp_entity = entity
            if project.type == constants.ProjectType.VECTOR.value:
                entity.segmentation_status = None
            if project.upload_state == constants.UploadState.EXTERNAL.value:
                tmp_entity.prediction_status = None
                tmp_entity.segmentation_status = None
            return TmpImageEntity(**tmp_entity.dict(by_alias=True))
        elif project.type == constants.ProjectType.VIDEO.value:
            return VideoEntity(**entity.dict(by_alias=True))
        elif project.type == constants.ProjectType.DOCUMENT.value:
            return DocumentEntity(**entity.dict(by_alias=True))
        return entity

    def execute(self) -> Response:
        if self.is_valid():
            condition = (
                Condition("name", self._item_name, EQ)
                & Condition("team_id", self._project.team_id, EQ)
                & Condition("project_id", self._project.id, EQ)
                & Condition("folder_id", self._folder.uuid, EQ)
                & Condition("includeCustomMetadata", self._include_custom_metadata, EQ)
            )
            response = self._backend_client.list_items(condition.build_query())
            if not response.ok:
                self._response.errors = response.error
                return self._response
            entity = next((i for i in response.data if i.name == self._item_name), None)
            if entity:
                entity = self.serialize_entity(entity, self._project)
                entity.add_path(self._project.name, self._folder.name)
                self._response.data = entity
            else:
                self._response.errors = AppException("Item not found.")
        return self._response


class QueryEntitiesUseCase(BaseReportableUseCase):
    def __init__(
        self,
        reporter: Reporter,
        project: ProjectEntity,
        folder: FolderEntity,
        backend_service_provider: SuperannotateServiceProvider,
        query: str,
        subset: str,
    ):
        super().__init__(reporter)
        self._project = project
        self._folder = folder
        self._backend_client = backend_service_provider
        self._query = query
        self._subset = subset

    def validate_arguments(self):
        if self._query:
            response = self._backend_client.validate_saqul_query(
                self._project.team_id, self._project.id, self._query
            )
            if response.get("error"):
                raise AppException(response["error"])
            if response["isValidQuery"]:
                self._query = response["parsedQuery"]
            else:
                raise AppException("Incorrect query.")
        else:
            response = self._backend_client.validate_saqul_query(
                self._project.team_id, self._project.id, "-"
            )
            if response.get("error"):
                raise AppException(response["error"])

        if not any([self._query, self._subset]):
            raise AppException(
                "The query and subset params cannot have the value None at the same time."
            )
        if self._subset and not self._folder.is_root:
            raise AppException(
                "The folder name should be specified in the query string."
            )

    def execute(self) -> Response:
        if self.is_valid():
            query_kwargs = {}
            if self._subset:
                subset: Optional[SubSetEntity] = None
                response = self._backend_client.list_sub_sets(
                    team_id=self._project.team_id, project_id=self._project.id
                )
                if response.ok:
                    subset = next(
                        (_sub for _sub in response.data if _sub.name == self._subset),
                        None,
                    )
                if not subset:
                    self._response.errors = AppException(
                        "Subset not found. Use the superannotate."
                        "get_subsets() function to get a list of the available subsets."
                    )
                    return self._response
                query_kwargs["subset_id"] = subset.id
            if self._query:
                query_kwargs["query"] = self._query
            query_kwargs["folder_id"] = (
                None if self._folder.name == "root" else self._folder.uuid
            )
            service_response = self._backend_client.saqul_query(
                self._project.team_id,
                self._project.id,
                **query_kwargs,
            )
            if service_response.ok:
                data = []
                for i, item in enumerate(service_response.data):
                    tmp_item = GetItem.serialize_entity(BaseItemEntity(**item), self._project)
                    folder_path = f"{'/' + item['folder_name'] if not item['is_root_folder'] else ''}"
                    tmp_item.path = f"{self._project.name}" + folder_path
                    data.append(tmp_item)
                self._response.data = data
            else:
                self._response.errors = service_response.data
        return self._response


class ListItems(BaseReportableUseCase):
    def __init__(
        self,
        reporter: Reporter,
        project: ProjectEntity,
        folder: FolderEntity,
        folders: BaseReadOnlyRepository,
        search_condition: Condition,
        backend_client: SuperannotateServiceProvider,
        recursive: bool = False,
        include_custom_metadata: bool = False,
    ):
        super().__init__(reporter)
        self._project = project
        self._folders = folders
        self._folder = folder
        self._backend_client = backend_client
        self._search_condition = search_condition
        self._recursive = recursive
        self._include_custom_metadata = include_custom_metadata

    def validate_recursive_case(self):
        if not self._folder.is_root and self._recursive:
            self._recursive = False

    def validate_project_type(self):
        if (
            self._project.type == constants.ProjectType.PIXEL.value
            and self._include_custom_metadata
        ):
            raise AppException(constants.METADATA_DEPRICATED_FOR_PIXEL)

    def execute(self) -> Response:
        if self.is_valid():
            self._search_condition &= Condition("team_id", self._project.team_id, EQ)
            self._search_condition &= Condition("project_id", self._project.id, EQ)
            self._search_condition &= Condition(
                "includeCustomMetadata", self._include_custom_metadata, EQ
            )

            if not self._recursive:
                self._search_condition &= Condition("folder_id", self._folder.uuid, EQ)
                items_response = self._backend_client.list_items(
                    self._search_condition.build_query()
                )
                if not items_response.ok:
                    raise AppException(items_response.error)
                items = []
                for item in items_response.data:
                    item = GetItem.serialize_entity(item, self._project)
                    item.add_path(self._project.name, self._folder.name)
                    items.append(item)
            else:
                items = []
                folders = self._folders.get_all(
                    Condition("team_id", self._project.team_id, EQ)
                    & Condition("project_id", self._project.id, EQ),
                )
                folders.append(self._folder)
                for folder in folders:
                    response = self._backend_client.list_items(
                        (
                            copy.deepcopy(self._search_condition)
                            & Condition("folder_id", folder.uuid, EQ)
                        ).build_query()
                    )
                    if not response.ok:
                        raise AppException(response.error)
                    for item in response.data:
                        item = GetItem.serialize_entity(item, self._project)
                        item.add_path(self._project.name, folder.name)
                        items.append(item)
            self._response.data = items
        return self._response


class AssignItemsUseCase(BaseUseCase):
    CHUNK_SIZE = 500

    def __init__(
        self,
        service: SuperannotateServiceProvider,
        project: ProjectEntity,
        folder: FolderEntity,
        item_names: list,
        user: str,
    ):
        super().__init__()
        self._project = project
        self._folder = folder
        self._item_names = item_names
        self._user = user
        self._service = service

    def validate_item_names(
        self,
    ):
        self._item_names = list(set(self._item_names))

    def execute(self):
        cnt_assigned = 0
        total_count = len(self._item_names)
        if self.is_valid():
            for i in range(0, len(self._item_names), self.CHUNK_SIZE):
                response = self._service.assign_items(
                    team_id=self._project.team_id,
                    project_id=self._project.id,
                    folder_name=self._folder.name,
                    user=self._user,
                    item_names=self._item_names[i : i + self.CHUNK_SIZE],  # noqa: E203
                )
                if not response.ok and response.error:  # User not found
                    self._response.errors += response.error
                    return self._response

                cnt_assigned += response.data["successCount"]
            logger.info(
                f"Assigned {cnt_assigned}/{total_count} items to user {self._user}"
            )
        return self._response


class UnAssignItemsUseCase(BaseUseCase):
    CHUNK_SIZE = 500

    def __init__(
        self,
        service: SuperannotateServiceProvider,
        project_entity: ProjectEntity,
        folder: FolderEntity,
        item_names: list,
    ):
        super().__init__()
        self._project_entity = project_entity
        self._folder = folder
        self._item_names = item_names
        self._service = service

    def execute(self):
        # todo handling to backend side
        for i in range(0, len(self._item_names), self.CHUNK_SIZE):
            is_un_assigned = self._service.un_assign_items(
                team_id=self._project_entity.team_id,
                project_id=self._project_entity.id,
                folder_name=self._folder.name,
                item_names=self._item_names[i : i + self.CHUNK_SIZE],  # noqa: E203
            )
            if not is_un_assigned:
                self._response.errors = AppException(
                    f"Cant un assign {', '.join(self._item_names[i: i + self.CHUNK_SIZE])}"
                )

        return self._response


class AttachItems(BaseReportableUseCase):
    CHUNK_SIZE = 500

    def __init__(
        self,
        reporter: Reporter,
        project: ProjectEntity,
        folder: FolderEntity,
        attachments: List[AttachmentEntity],
        annotation_status: str,
        backend_service_provider: SuperannotateServiceProvider,
        upload_state_code: int = constants.UploadState.EXTERNAL.value,
    ):
        super().__init__(reporter)
        self._project = project
        self._folder = folder
        self._attachments = attachments
        self._annotation_status_code = constants.AnnotationStatus.get_value(
            annotation_status
        )
        self._upload_state_code = upload_state_code
        self._backend_service = backend_service_provider
        self._attachments_count = None

    @property
    def attachments_count(self):
        if not self._attachments_count:
            self._attachments_count = len(self._attachments)
        return self._attachments_count

    def validate_limitations(self):
        attachments_count = self.attachments_count
        response = self._backend_service.get_limitations(
            team_id=self._project.team_id,
            project_id=self._project.id,
            folder_id=self._folder.uuid,
        )
        if not response.ok:
            raise AppValidationException(response.error)
        if attachments_count > response.data.folder_limit.remaining_image_count:
            raise AppValidationException(constants.ATTACH_FOLDER_LIMIT_ERROR_MESSAGE)
        elif attachments_count > response.data.project_limit.remaining_image_count:
            raise AppValidationException(constants.ATTACH_PROJECT_LIMIT_ERROR_MESSAGE)
        elif (
            response.data.user_limit
            and attachments_count > response.data.user_limit.remaining_image_count
        ):
            raise AppValidationException(constants.ATTACH_USER_LIMIT_ERROR_MESSAGE)

    def validate_upload_state(self):
        if self._project.upload_state == constants.UploadState.BASIC.value:
            raise AppValidationException(constants.ATTACHING_UPLOAD_STATE_ERROR)

    @staticmethod
    def generate_meta():
        return {"width": None, "height": None}

    def execute(self) -> Response:
        if self.is_valid():
            duplications = []
            attached = []
            self.reporter.start_progress(self.attachments_count, "Attaching URLs")
            for i in range(0, self.attachments_count, self.CHUNK_SIZE):
                attachments = self._attachments[i : i + self.CHUNK_SIZE]  # noqa: E203
                response = self._backend_service.get_bulk_images(
                    project_id=self._project.id,
                    team_id=self._project.team_id,
                    folder_id=self._folder.uuid,
                    images=[attachment.name for attachment in attachments],
                )
                if isinstance(response, dict) and "error" in response:
                    raise AppException(response["error"])
                duplications.extend([image["name"] for image in response])
                to_upload = []
                to_upload_meta = {}
                for attachment in attachments:
                    if attachment.name not in duplications:
                        to_upload.append(
                            {"name": attachment.name, "path": attachment.url}
                        )
                        to_upload_meta[attachment.name] = self.generate_meta()
                if to_upload:
                    backend_response = self._backend_service.attach_files(
                        project_id=self._project.id,
                        folder_id=self._folder.uuid,
                        team_id=self._project.team_id,
                        files=to_upload,
                        annotation_status_code=self._annotation_status_code,
                        upload_state_code=self._upload_state_code,
                        meta=to_upload_meta,
                    )
                    if "error" in backend_response:
                        self._response.errors = AppException(backend_response["error"])
                    else:
                        attached.extend(backend_response)
                self.reporter.update_progress(len(attachments))
            self.reporter.finish_progress()
            self._response.data = attached, duplications
        return self._response


class CopyItems(BaseReportableUseCase):
    """
    Copy items in bulk between folders in a project.
    Return skipped item names.
    """

    CHUNK_SIZE = 500

    def __init__(
        self,
        reporter: Reporter,
        project: ProjectEntity,
        from_folder: FolderEntity,
        to_folder: FolderEntity,
        item_names: List[str],
        items: BaseReadOnlyRepository,
        backend_service_provider: SuperannotateServiceProvider,
        include_annotations: bool,
    ):
        super().__init__(reporter)
        self._project = project
        self._from_folder = from_folder
        self._to_folder = to_folder
        self._item_names = item_names
        self._items = items
        self._backend_service = backend_service_provider
        self._include_annotations = include_annotations

    def _validate_limitations(self, items_count):
        response = self._backend_service.get_limitations(
            team_id=self._project.team_id,
            project_id=self._project.id,
            folder_id=self._to_folder.uuid,
        )
        if not response.ok:
            raise AppValidationException(response.error)
        if items_count > response.data.folder_limit.remaining_image_count:
            raise AppValidationException(constants.COPY_FOLDER_LIMIT_ERROR_MESSAGE)
        if items_count > response.data.project_limit.remaining_image_count:
            raise AppValidationException(constants.COPY_PROJECT_LIMIT_ERROR_MESSAGE)

    def validate_item_names(self):
        if self._item_names:
            self._item_names = list(set(self._item_names))

    def execute(self):
        if self.is_valid():
            if self._item_names:
                items = self._item_names
            else:
                condition = (
                    Condition("team_id", self._project.team_id, EQ)
                    & Condition("project_id", self._project.id, EQ)
                    & Condition("folder_id", self._from_folder.uuid, EQ)
                )
                items = [item.name for item in self._items.get_all(condition)]

            existing_items = []
            for i in range(0, len(items), self.CHUNK_SIZE):
                cand_items = self._backend_service.get_bulk_images(
                    project_id=self._project.id,
                    team_id=self._project.team_id,
                    folder_id=self._to_folder.uuid,
                    images=items[i : i + self.CHUNK_SIZE],
                )
                if isinstance(cand_items, dict):
                    continue
                existing_items += cand_items

            duplications = [item["name"] for item in existing_items]
            items_to_copy = list(set(items) - set(duplications))
            skipped_items = duplications
            try:
                self._validate_limitations(len(items_to_copy))
            except AppValidationException as e:
                self._response.errors = e
                return self._response
            if items_to_copy:
                for i in range(0, len(items_to_copy), self.CHUNK_SIZE):
                    chunk_to_copy = items_to_copy[i : i + self.CHUNK_SIZE]  # noqa: E203
                    poll_id = (
                        self._backend_service.copy_items_between_folders_transaction(
                            team_id=self._project.team_id,
                            project_id=self._project.id,
                            from_folder_id=self._from_folder.uuid,
                            to_folder_id=self._to_folder.uuid,
                            items=chunk_to_copy,
                            include_annotations=self._include_annotations,
                        )
                    )
                    if not poll_id:
                        skipped_items.extend(chunk_to_copy)
                        continue
                    try:
                        self._backend_service.await_progress(
                            self._project.id,
                            self._project.team_id,
                            poll_id=poll_id,
                            items_count=len(chunk_to_copy),
                        )
                    except BackendError as e:
                        self._response.errors = AppException(e)
                        return self._response

                existing_items = []
                for i in range(0, len(items), self.CHUNK_SIZE):
                    cand_items = self._backend_service.get_bulk_images(
                        project_id=self._project.id,
                        team_id=self._project.team_id,
                        folder_id=self._to_folder.uuid,
                        images=items[i : i + self.CHUNK_SIZE],
                    )
                    if isinstance(cand_items, dict):
                        continue
                    existing_items += cand_items

                existing_item_names_set = {item["name"] for item in existing_items}
                items_to_copy_names_set = set(items_to_copy)
                copied_items = existing_item_names_set.intersection(
                    items_to_copy_names_set
                )
                skipped_items.extend(list(items_to_copy_names_set - copied_items))
                self.reporter.log_info(
                    f"Copied {len(copied_items)}/{len(items)} item(s) from "
                    f"{self._project.name}{'' if self._from_folder.is_root else f'/{self._from_folder.name}'} to "
                    f"{self._project.name}{'' if self._to_folder.is_root else f'/{self._to_folder.name}'}"
                )
            self._response.data = skipped_items
        return self._response


class MoveItems(BaseReportableUseCase):
    CHUNK_SIZE = 1000

    def __init__(
        self,
        reporter: Reporter,
        project: ProjectEntity,
        from_folder: FolderEntity,
        to_folder: FolderEntity,
        item_names: List[str],
        items: BaseReadOnlyRepository,
        backend_service_provider: SuperannotateServiceProvider,
    ):
        super().__init__(reporter)
        self._project = project
        self._from_folder = from_folder
        self._to_folder = to_folder
        self._item_names = item_names
        self._items = items
        self._backend_service = backend_service_provider

    def validate_item_names(self):
        if self._item_names:
            self._item_names = list(set(self._item_names))

    def _validate_limitations(self, items_count):
        response = self._backend_service.get_limitations(
            team_id=self._project.team_id,
            project_id=self._project.id,
            folder_id=self._to_folder.uuid,
        )
        if not response.ok:
            raise AppValidationException(response.error)
        if items_count > response.data.folder_limit.remaining_image_count:
            raise AppValidationException(constants.MOVE_FOLDER_LIMIT_ERROR_MESSAGE)
        if items_count > response.data.project_limit.remaining_image_count:
            raise AppValidationException(constants.MOVE_PROJECT_LIMIT_ERROR_MESSAGE)

    def execute(self):
        if self.is_valid():
            if not self._item_names:
                condition = (
                    Condition("team_id", self._project.team_id, EQ)
                    & Condition("project_id", self._project.id, EQ)
                    & Condition("folder_id", self._from_folder.uuid, EQ)
                )
                items = [item.name for item in self._items.get_all(condition)]
            else:
                items = self._item_names
            try:
                self._validate_limitations(len(items))
            except AppValidationException as e:
                self._response.errors = e
                return self._response
            moved_images = []
            for i in range(0, len(items), self.CHUNK_SIZE):
                moved_images.extend(
                    self._backend_service.move_images_between_folders(
                        team_id=self._project.team_id,
                        project_id=self._project.id,
                        from_folder_id=self._from_folder.uuid,
                        to_folder_id=self._to_folder.uuid,
                        images=items[i : i + self.CHUNK_SIZE],  # noqa: E203
                    )
                )
            self.reporter.log_info(
                f"Moved {len(moved_images)}/{len(items)} item(s) from "
                f"{self._project.name}{'' if self._from_folder.is_root else f'/{self._from_folder.name}'} to "
                f"{self._project.name}{'' if self._to_folder.is_root else f'/{self._to_folder.name}'}"
            )

            self._response.data = list(set(items) - set(moved_images))
        return self._response


class SetAnnotationStatues(BaseReportableUseCase):
    CHUNK_SIZE = 500
    ERROR_MESSAGE = "Failed to change status"

    def __init__(
        self,
        reporter: Reporter,
        project: ProjectEntity,
        folder: FolderEntity,
        items: BaseReadOnlyRepository,
        annotation_status: str,
        backend_service_provider: SuperannotateServiceProvider,
        item_names: List[str] = None,
    ):
        super().__init__(reporter)
        self._project = project
        self._folder = folder
        self._item_names = item_names
        self._items = items
        self._annotation_status_code = constants.AnnotationStatus.get_value(
            annotation_status
        )
        self._backend_service = backend_service_provider

    def validate_items(self):
        if not self._item_names:
            condition = (
                Condition("team_id", self._project.team_id, EQ)
                & Condition("project_id", self._project.id, EQ)
                & Condition("folder_id", self._folder.uuid, EQ)
            )
            self._item_names = [item.name for item in self._items.get_all(condition)]
            return
        existing_items = []
        for i in range(0, len(self._item_names), self.CHUNK_SIZE):

            search_names = self._item_names[i : i + self.CHUNK_SIZE]
            cand_items = self._backend_service.get_bulk_images(
                project_id=self._project.id,
                team_id=self._project.team_id,
                folder_id=self._folder.uuid,
                images=search_names,
            )

            if isinstance(cand_items, dict):
                continue
            existing_items += cand_items

        if not existing_items:
            raise AppValidationException(self.ERROR_MESSAGE)
        if existing_items:
            self._item_names = list(
                {i["name"] for i in existing_items}.intersection(set(self._item_names))
            )

    def execute(self):
        if self.is_valid():
            for i in range(0, len(self._item_names), self.CHUNK_SIZE):
                status_changed = self._backend_service.set_images_statuses_bulk(
                    image_names=self._item_names[
                        i : i + self.CHUNK_SIZE
                    ],  # noqa: E203,
                    team_id=self._project.team_id,
                    project_id=self._project.id,
                    folder_id=self._folder.uuid,
                    annotation_status=self._annotation_status_code,
                )
                if not status_changed:
                    self._response.errors = AppException(self.ERROR_MESSAGE)
                    break
        return self._response


class DeleteItemsUseCase(BaseUseCase):
    CHUNK_SIZE = 1000

    def __init__(
        self,
        project: ProjectEntity,
        folder: FolderEntity,
        backend_service_provider: SuperannotateServiceProvider,
        items: BaseReadOnlyRepository,
        item_names: List[str] = None,
    ):
        super().__init__()
        self._project = project
        self._folder = folder
        self._items = items
        self._backend_service = backend_service_provider
        self._item_names = item_names

    def execute(self):
        if self.is_valid():
            if self._item_names:
                item_ids = [
                    item.uuid
                    for item in GetBulkItems(
                        service=self._backend_service,
                        project_id=self._project.id,
                        team_id=self._project.team_id,
                        folder_id=self._folder.uuid,
                        items=self._item_names,
                    )
                    .execute()
                    .data
                ]
            else:
                condition = (
                    Condition("team_id", self._project.team_id, EQ)
                    & Condition("project_id", self._project.id, EQ)
                    & Condition("folder_id", self._folder.uuid, EQ)
                )
                item_ids = [item.id for item in self._items.get_all(condition)]

            for i in range(0, len(item_ids), self.CHUNK_SIZE):
                response = self._backend_service.delete_items(
                    project_id=self._project.id,
                    team_id=self._project.team_id,
                    item_ids=item_ids[i : i + self.CHUNK_SIZE],  # noqa: E203
                )

            logger.info(
                f"Items deleted in project {self._project.name}{'/' + self._folder.name if not self._folder.is_root else ''}"
            )

        return self._response



from typing import Optional

from lib.core.entities.base import BaseItemEntity
from lib.core.enums import ApprovalStatus
from lib.core.enums import SegmentationStatus
from pydantic import Extra
from pydantic import Field


class TmpImageEntity(BaseItemEntity):
    prediction_status: Optional[SegmentationStatus] = Field(
        SegmentationStatus.NOT_STARTED
    )
    segmentation_status: Optional[SegmentationStatus] = Field(
        SegmentationStatus.NOT_STARTED
    )
    approval_status: Optional[ApprovalStatus] = Field(None)
    is_pinned: Optional[bool]

    class Config:
        extra = Extra.ignore


class VideoEntity(BaseItemEntity):
    approval_status: Optional[ApprovalStatus] = Field(None)

    class Config:
        extra = Extra.ignore


class DocumentEntity(BaseItemEntity):
    class Config:
        extra = Extra.ignore



from .conversion import convert_json_version
from .conversion import convert_project_type
from .conversion import export_annotation
from .conversion import import_annotation


__all__ = [
    "convert_json_version",
    "convert_project_type",
    "export_annotation",
    "import_annotation",
]



import json

import numpy as np
from superannotate.logger import get_default_logger
from tqdm import tqdm

logger = get_default_logger()


def hex_to_rgb(hex_string):
    """Converts HEX values to RGB values"""
    h = hex_string.lstrip("#")
    return tuple(int(h[i : i + 2], 16) for i in (0, 2, 4))


def blue_color_generator(n, hex_values=True):
    """Blue colors generator for SuperAnnotate blue mask."""
    hex_colors = []
    for i in range(n + 1):
        int_color = i * 15
        bgr_color = np.array(
            [int_color & 255, (int_color >> 8) & 255, (int_color >> 16) & 255, 255],
            dtype=np.uint8,
        )
        hex_color = (
            "#"
            + "{:02x}".format(bgr_color[2])
            + "{:02x}".format(
                bgr_color[1],
            )
            + "{:02x}".format(bgr_color[0])
        )
        if hex_values:
            hex_colors.append(hex_color)
        else:
            hex_colors.append(hex_to_rgb(hex_color))
    return hex_colors[1:]


def id2rgb(id_map):
    if isinstance(id_map, np.ndarray):
        id_map_copy = id_map.copy()
        rgb_shape = tuple(list(id_map.shape) + [3])
        rgb_map = np.zeros(rgb_shape, dtype=np.uint8)
        for i in range(3):
            rgb_map[..., i] = id_map_copy % 256
            id_map_copy //= 256
        return rgb_map
    color = []
    for _ in range(3):
        color.append(id_map % 256)
        id_map //= 256
    return color


def write_to_json(output_path, json_data):
    with open(output_path, "w") as fw:
        json.dump(json_data, fw, indent=2)


MAX_IMAGE_SIZE = 100 * 1024 * 1024  # 100 MB limit


def tqdm_converter(total_num, images_converted, images_not_converted, finish_event):
    with tqdm(total=total_num) as pbar:
        while True:
            finished = finish_event.wait(5)
            if not finished:
                sum_all = len(images_converted) + len(images_not_converted)
                pbar.update(sum_all - pbar.n)
            else:
                pbar.update(total_num - pbar.n)
                break


def get_annotation_json_name(image_name, project_type):
    if project_type == "Vector":
        return image_name + "___objects.json"
    else:
        return image_name + "___pixel.json"


def get_annotation_png_name(image_name):
    return image_name + "___save.png"



import random
import time
from abc import ABCMeta
from abc import abstractmethod
from collections import defaultdict
from functools import lru_cache
from typing import Any
from typing import Callable
from typing import Dict
from typing import List

import lib.core as constances
from lib.core.enums import AnnotationTypes
from lib.core.reporter import Reporter
from superannotate_schemas.schemas.classes import AnnotationClass
from superannotate_schemas.schemas.classes import Attribute
from superannotate_schemas.schemas.classes import AttributeGroup


class BaseDataHandler(metaclass=ABCMeta):
    @abstractmethod
    def handle(self, *args, **kwargs):
        raise NotImplementedError


class ClassIdGenerator:
    def __init__(self):
        self.classes = defaultdict(int)
        self.idx = -1

    def send(self, class_name: str):
        if class_name not in self.classes:
            self.classes[class_name] = self.idx
            self.idx -= 1
        return self.classes[class_name]


class BaseAnnotationDateHandler(BaseDataHandler, metaclass=ABCMeta):
    def __init__(self, annotation_classes: List[AnnotationClass]):
        self._annotation_classes: List[AnnotationClass] = annotation_classes

    @lru_cache()
    def get_annotation_class(self, name: str) -> AnnotationClass:
        for annotation_class in self._annotation_classes:
            if annotation_class.name == name:
                return annotation_class

    @lru_cache()
    def get_attribute_group(
        self, annotation_class: AnnotationClass, attr_group_name: str
    ) -> AttributeGroup:
        for attr_group in annotation_class.attribute_groups:
            if attr_group.name == attr_group_name:
                return attr_group

    @lru_cache()
    def get_attribute(self, attr_group: AttributeGroup, attr_name: str) -> Attribute:
        for attr in attr_group.attributes:
            if attr.name == attr_name:
                return attr


class ChainedAnnotationHandlers:
    def __init__(self):
        self._handlers: List[BaseDataHandler] = list()

    def handle(self, data: Any):
        for handler in self._handlers:
            data = handler.handle(data)
        return data

    def attach(self, handler: BaseDataHandler):
        if handler not in self._handlers:
            self._handlers.append(handler)

    def detach(self, handler):
        if handler in self._handlers:
            self._handlers.remove(handler)

    def detach_all(self):
        self._handlers.clear()


class AnnotationStatusHandler(BaseDataHandler):
    def handle(self, annotations: Dict):
        annotations["metadata"][
            "status"
        ] = constances.AnnotationStatus.IN_PROGRESS.value
        return annotations


class LastActionHandler(BaseDataHandler):
    def __init__(self, email: str):
        self._email = email

    def handle(self, annotations: Dict):
        annotations["metadata"]["lastAction"] = {
            "email": self._email,
            "timestamp": int(time.time()),
        }
        return annotations


class DocumentTagHandler(BaseAnnotationDateHandler):
    def handle(self, annotation: dict):
        new_tags = []
        for tag in annotation["tags"]:
            annotation_class = self.get_annotation_class(tag)
            if annotation_class:
                new_tags.append(annotation_class.id)
        annotation["tags"] = new_tags
        return annotation


class MissingIDsHandler(BaseAnnotationDateHandler):
    def __init__(
        self,
        annotation_classes: List[AnnotationClass],
        templates: List[dict],
        reporter: Reporter,
    ):
        super().__init__(annotation_classes)
        self.validate_existing_classes(annotation_classes)
        self._templates = templates
        self.reporter = reporter

    def validate_existing_classes(self, annotation_classes: List[AnnotationClass]):
        classes_data = defaultdict(dict)
        for annotation_class in annotation_classes:
            class_info = {"id": annotation_class.id, "attribute_groups": {}}
            if annotation_class.attribute_groups:
                for attribute_group in annotation_class.attribute_groups:
                    attribute_group_data = defaultdict(dict)
                    for attribute in attribute_group.attributes:
                        if attribute.name in attribute_group_data.keys():
                            self.reporter.log_warning(
                                f"Duplicate annotation class attribute name {attribute.name}"
                                f" in attribute group {attribute_group.name}. "
                                "Only one of the annotation class attributes will be used. "
                                "This will result in errors in annotation upload."
                            )
                        attribute_group_data[attribute.name] = attribute.id
                    if attribute_group.name in class_info["attribute_groups"].keys():
                        self.reporter.log_warning(
                            f"Duplicate annotation class attribute group name {attribute_group.name}."
                            " Only one of the annotation class attribute groups will be used."
                            " This will result in errors in annotation upload."
                        )
                    class_info["attribute_groups"][attribute_group.name] = {
                        "id": attribute_group.id,
                        "attributes": attribute_group_data,
                    }
            if annotation_class.name in classes_data.keys():
                self.reporter.log_warning(
                    f"Duplicate annotation class name {annotation_class.name}."
                    f" Only one of the annotation classes will be used."
                    " This will result in errors in annotation upload.",
                )

    def handle(self, annotation: dict):
        if "instances" not in annotation:
            return annotation
        for annotation_instance in annotation["instances"]:
            if "className" not in annotation_instance:
                annotation_instance["classId"] = -1
            else:
                class_name = annotation_instance["className"]
                annotation_class = self.get_annotation_class(class_name)
                if not annotation_class:
                    self.reporter.log_warning(f"Couldn't find class {class_name}")
                    self.reporter.store_message("missing_classes", class_name)
                    annotation_instance["classId"] = -1
                    annotation_instance["attributes"] = []
                    self._annotation_classes.append(
                        AnnotationClass(
                            id=-1,
                            name=class_name,
                            attribute_groups=[],
                            color=f"#{random.randint(0, 0xFFFFFF):06x}",
                        )
                    )
                    self.get_annotation_class.cache_clear()
                else:
                    annotation_instance["classId"] = annotation_class.id

        template_name_id_map = {
            template["name"]: template["id"] for template in self._templates
        }
        for annotation_instance in (
            i for i in annotation["instances"] if i.get("type", None) == "template"
        ):
            annotation_instance["templateId"] = template_name_id_map.get(
                annotation_instance.get("templateName", ""), -1
            )

        for annotation_instance in [
            i for i in annotation["instances"] if "className" in i and i["classId"] > 0
        ]:
            annotation_class_name = annotation_instance["className"]
            annotation_class = self.get_annotation_class(annotation_class_name)

            if not annotation_class:
                self.reporter.log_warning(
                    f"Couldn't find annotation class {annotation_class_name}"
                )
                continue
            annotation_instance_attributes = []
            for annotation_attribute in annotation_instance["attributes"]:
                attr_group = self.get_attribute_group(
                    annotation_class, annotation_attribute["groupName"]
                )
                if not attr_group:
                    self.reporter.log_warning(
                        f"Couldn't find annotation group {annotation_attribute['groupName']}."
                    )
                    self.reporter.store_message(
                        "missing_attribute_groups",
                        f"{annotation_instance['className']}.{annotation_attribute['groupName']}",
                    )
                    continue
                annotation_attribute["groupId"] = attr_group.id
                attribute = self.get_attribute(attr_group, annotation_attribute["name"])
                if not attribute:
                    del annotation_attribute["groupId"]
                    self.reporter.log_warning(
                        f"Couldn't find annotation name {annotation_attribute['name']} in"
                        f" annotation group {annotation_attribute['groupName']}",
                    )
                    self.reporter.store_message(
                        "missing_attributes", annotation_attribute["name"]
                    )
                    continue
                annotation_attribute["id"] = attribute.id
                annotation_instance_attributes.append(annotation_attribute)
            annotation_instance["attributes"] = annotation_instance_attributes
        return annotation


class VideoFormatHandler(BaseAnnotationDateHandler):
    INSTANCE_FIELDS = {
        "pointLabels",
        "createdBy",
        "createdAt",
        "updatedBy",
        "updatedAt",
    }

    @staticmethod
    def _point_handler(time_stamp):
        pass

    HANDLERS: Dict[str, Callable] = {
        AnnotationTypes.EVENT: lambda timestamp: {},
        AnnotationTypes.BBOX: lambda timestamp: {"points": timestamp["points"]},
        AnnotationTypes.POINT: lambda timestamp: {
            "x": timestamp["x"],
            "y": timestamp["y"],
        },
        AnnotationTypes.POLYLINE: lambda timestamp: {"points": timestamp["points"]},
        AnnotationTypes.POLYGON: lambda timestamp: {
            "points": timestamp["points"],
            "exclude": timestamp.get("exclude", []),
        },
    }

    def __init__(self, annotation_classes: List[AnnotationClass], reporter: Reporter):
        super().__init__(annotation_classes)
        self.reporter = reporter

    def handle(self, annotation: dict):
        id_generator = ClassIdGenerator()

        def safe_time(timestamp):
            return "0" if str(timestamp) == "0.0" else timestamp

        def convert_timestamp(timestamp):
            return timestamp / 10**6 if timestamp else "0"

        editor_data = {
            "instances": [],
            "tags": annotation["tags"],
            "name": annotation["metadata"]["name"],
            "metadata": {
                "name": annotation["metadata"]["name"],
                "width": annotation["metadata"].get("width"),
                "height": annotation["metadata"].get("height"),
            },
        }
        if annotation["metadata"].get("duration"):
            editor_data["metadata"]["duration"] = convert_timestamp(
                annotation["metadata"]["duration"]
            )
        for instance in annotation["instances"]:
            meta = instance["meta"]
            class_name = meta.get("className")
            editor_instance = {
                "attributes": [],
                "timeline": {},
                "type": meta["type"],
                "locked": False,
            }
            if class_name:
                annotation_class = self.get_annotation_class(class_name)
                if annotation_class:
                    editor_instance["classId"] = annotation_class.id
                else:
                    editor_instance["classId"] = id_generator.send(class_name)
            else:
                editor_instance["classId"] = id_generator.send("unknown_class")

            matched_fields = self.INSTANCE_FIELDS & meta.keys()
            for matched_field in matched_fields:
                editor_instance[matched_field] = meta[matched_field]
            active_attributes = set()
            for parameter in instance["parameters"]:
                start_time = safe_time(convert_timestamp(parameter["start"]))
                end_time = safe_time(convert_timestamp(parameter["end"]))
                for timestamp_data in parameter["timestamps"]:
                    timestamp = safe_time(
                        convert_timestamp(timestamp_data["timestamp"])
                    )
                    editor_instance["timeline"][timestamp] = {}

                    if timestamp == start_time:
                        editor_instance["timeline"][timestamp]["active"] = True

                    if timestamp == end_time:
                        editor_instance["timeline"][timestamp]["active"] = False
                    handler: Callable = self.HANDLERS.get(meta["type"])
                    if handler:
                        editor_instance["timeline"][timestamp].update(
                            handler(timestamp_data)
                        )
                    # if timestamp_data.get("points", None):
                    #     editor_instance["timeline"][timestamp][
                    #         "points"
                    #     ] = timestamp_data["points"]
                    if not class_name:
                        continue
                    annotation_class = self.get_annotation_class(class_name)
                    if not annotation_class:
                        self.reporter.store_message(
                            "missing_classes", meta["className"]
                        )
                        continue

                    existing_attributes_in_current_instance = set()
                    for attribute in timestamp_data["attributes"]:
                        group_name, attr_name = (
                            attribute.get("groupName"),
                            attribute.get("name"),
                        )
                        attr_group = self.get_attribute_group(
                            annotation_class, group_name
                        )
                        attribute = (
                            self.get_attribute(attr_group, attr_name)
                            if attr_group
                            else None
                        )
                        if not attr_group:
                            self.reporter.store_message(
                                "missing_attribute_groups", f"{class_name}.{group_name}"
                            )
                        elif not attribute:
                            self.reporter.store_message(
                                "missing_attributes",
                                f"{class_name}.{group_name}.{attr_name}",
                            )
                        else:
                            existing_attributes_in_current_instance.add(
                                (group_name, attr_name)
                            )
                    attributes_to_add = (
                        existing_attributes_in_current_instance - active_attributes
                    )
                    attributes_to_delete = (
                        active_attributes - existing_attributes_in_current_instance
                    )
                    if attributes_to_add or attributes_to_delete:
                        editor_instance["timeline"][timestamp][
                            "attributes"
                        ] = defaultdict(list)
                    for new_attribute in attributes_to_add:
                        attribute_group = self.get_attribute_group(
                            annotation_class, new_attribute[0]
                        )
                        attribute = self.get_attribute(
                            attribute_group, new_attribute[1]
                        )
                        attr = {
                            "id": attribute.id,
                            "groupId": attribute_group.id,
                        }
                        active_attributes.add(new_attribute)
                        editor_instance["timeline"][timestamp]["attributes"][
                            "+"
                        ].append(attr)
                    for attribute_to_delete in attributes_to_delete:
                        attribute_group = self.get_attribute_group(
                            annotation_class, attribute_to_delete[0]
                        )
                        attribute = self.get_attribute(
                            attribute_group, attribute_to_delete[1]
                        )
                        attr = {
                            "id": attribute.id,
                            "groupId": attribute_group.id,
                        }
                        active_attributes.remove(attribute_to_delete)
                        editor_instance["timeline"][timestamp]["attributes"][
                            "-"
                        ].append(attr)
            editor_data["instances"].append(editor_instance)
        return editor_data



import threading
from typing import Dict

from dataclasses import dataclass
from dataclasses import field


class Session:
    def __init__(self):
        self.pk = threading.get_ident()
        self._data_dict = {}

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if type is not None:
            return False

    def __del__(self):
        Config().delete_current_session()

    @property
    def data(self):
        return self._data_dict

    @staticmethod
    def get_current_session():
        return Config().get_current_session()

    def __setitem__(self, key, item):
        self._data_dict[key] = item

    def __getitem__(self, key):
        return self._data_dict[key]

    def __repr__(self):
        return repr(self._data_dict)

    def clear(self):
        return self._data_dict.clear()


class Singleton(type):
    _instances = {}
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


@dataclass()
class Config(metaclass=Singleton):
    SESSIONS: Dict[int, Session] = field(default_factory=dict)

    def get_current_session(self):
        session = self.SESSIONS.get(threading.get_ident())
        if not session:
            session = Session()
            self.SESSIONS.update({session.pk: session})
        return session

    def delete_current_session(self):
        ident = threading.get_ident()
        if ident in self.SESSIONS:
            del self.SESSIONS[ident]



from abc import abstractmethod
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List
from typing import Tuple

from lib.core.reporter import Reporter
from lib.core.service_types import ServiceResponse


class SuperannotateServiceProvider:
    @abstractmethod
    def attach_files(
        self,
        project_id: int,
        folder_id: int,
        team_id: int,
        files: List[Dict],
        annotation_status_code: int,
        upload_state_code: int,
        meta: Dict,
    ):
        raise NotImplementedError

    @abstractmethod
    def get_annotation_classes(
        self, project_id: int, team_id: int, name_prefix: str = None
    ):
        raise NotImplementedError

    @abstractmethod
    def share_project_bulk(self, project_id: int, team_id: int, users: Iterable):
        raise NotImplementedError

    @abstractmethod
    def invite_contributors(
        self, team_id: int, team_role: int, emails: Iterable
    ) -> Tuple[List[str], List[str]]:
        raise NotImplementedError

    @abstractmethod
    def prepare_export(
        self,
        project_id: int,
        team_id: int,
        folders: List[str],
        annotation_statuses: Iterable[Any],
        include_fuse: bool,
        only_pinned: bool,
    ):
        raise NotImplementedError

    @abstractmethod
    def search_team_contributors(self, team_id: int, query_string: str = None):
        raise NotImplementedError

    @abstractmethod
    def get_project_settings(self, project_id: int, team_id: int):
        raise NotImplementedError

    @abstractmethod
    def set_project_settings(self, project_id: int, team_id: int, data: List):
        raise NotImplementedError

    @abstractmethod
    def get_project_workflows(self, project_id: int, team_id: int):
        raise NotImplementedError

    @abstractmethod
    def list_images(self, query_string):
        raise NotImplementedError

    @abstractmethod
    def get_project(self, uuid: int, team_id: int):
        raise NotImplementedError

    @abstractmethod
    def set_project_workflow(self, project_id: int, team_id: int, data: Dict):
        raise NotImplementedError

    def delete_folders(self, project_id: int, team_id: int, folder_ids: List[int]):
        raise NotImplementedError

    def get_folder(self, query_string: str):
        raise NotImplementedError

    def get_folders(self, query_string: str = None, params: dict = None):
        raise NotImplementedError

    def create_folder(self, project_id: int, team_id: int, folder_name: str):
        raise NotImplementedError

    def update_folder(self, project_id: int, team_id: int, folder_data: dict):
        raise NotImplementedError

    def list_items(self, query_params: str) -> ServiceResponse:
        raise NotImplementedError

    def get_download_token(
        self,
        project_id: int,
        team_id: int,
        folder_id: int,
        image_id: int,
        include_original: int = 1,
    ) -> dict:
        raise NotImplementedError

    def get_upload_token(
        self,
        project_id: int,
        team_id: int,
        folder_id: int,
        image_id: int,
    ) -> dict:
        raise NotImplementedError

    def update_image(self, image_id: int, team_id: int, project_id: int, data: dict):
        raise NotImplementedError

    def copy_items_between_folders_transaction(
        self,
        team_id: int,
        project_id: int,
        from_folder_id: int,
        to_folder_id: int,
        items: List[str],
        include_annotations: bool = False,
        include_pin: bool = False,
    ) -> int:
        raise NotImplementedError

    def move_images_between_folders(
        self,
        team_id: int,
        project_id: int,
        from_folder_id: int,
        to_folder_id: int,
        images: List[str],
    ) -> List[str]:
        """
        Returns list of moved images.
        """
        raise NotImplementedError

    def get_progress(
        self, project_id: int, team_id: int, poll_id: int
    ) -> Tuple[int, int]:
        raise NotImplementedError

    def await_progress(
        self, project_id: int, team_id: int, poll_id: int, items_count
    ) -> Tuple[int, int]:
        raise NotImplementedError

    def set_images_statuses_bulk(
        self,
        image_names: List[str],
        team_id: int,
        project_id: int,
        folder_id: int,
        annotation_status: int,
    ):
        raise NotImplementedError

    def delete_items(self, project_id: int, team_id: int, item_ids: List[int]):
        raise NotImplementedError

    def assign_items(
        self,
        team_id: int,
        project_id: int,
        folder_name: str,
        user: str,
        item_names: list,
    ) -> ServiceResponse:
        raise NotImplementedError

    def get_bulk_images(
        self, project_id: int, team_id: int, folder_id: int, images: List[str]
    ) -> List[dict]:
        raise NotImplementedError

    def un_assign_folder(
        self,
        team_id: int,
        project_id: int,
        folder_name: str,
    ):
        raise NotImplementedError

    def assign_folder(
        self, team_id: int, project_id: int, folder_name: str, users: list
    ):
        raise NotImplementedError

    def un_assign_items(
        self,
        team_id: int,
        project_id: int,
        folder_name: str,
        item_names: list,
    ):
        raise NotImplementedError

    def un_share_project(
        self,
        team_id: int,
        project_id: int,
        user_id: str,
    ):
        raise NotImplementedError

    def get_exports(self, team_id: int, project_id: int):
        raise NotImplementedError

    def get_export(self, team_id: int, project_id: int, export_id: int):
        raise NotImplementedError

    def get_project_images_count(self, team_id: int, project_id: int):
        raise NotImplementedError

    def get_s3_upload_auth_token(self, team_id: int, folder_id: int, project_id: int):
        raise NotImplementedError

    def delete_annotation_class(
        self, team_id: int, project_id: int, annotation_class_id: int
    ):
        raise NotImplementedError

    def set_annotation_classes(self, team_id: int, project_id: int, data: list):
        raise NotImplementedError

    def set_project_workflow_bulk(self, project_id: int, team_id: int, steps: list):
        raise NotImplementedError

    def set_project_workflow_attributes_bulk(
        self, project_id: int, team_id: int, attributes: list
    ):
        raise NotImplementedError

    def get_pre_annotation_upload_data(
        self, project_id: int, team_id: int, image_ids: List[int], folder_id: int
    ):
        raise NotImplementedError

    def get_annotation_upload_data(
        self, project_id: int, team_id: int, image_ids: List[int], folder_id: int
    ) -> ServiceResponse:
        raise NotImplementedError

    def get_templates(self, team_id: int):
        raise NotImplementedError

    def start_model_training(self, team_id: int, hyper_parameters: dict) -> dict:
        raise NotImplementedError

    def get_model_metrics(self, team_id: int, model_id: int) -> dict:
        raise NotImplementedError

    def bulk_get_folders(self, team_id: int, project_ids: List[int]):
        raise NotImplementedError

    def update_model(self, team_id: int, model_id: int, data: dict):
        raise NotImplementedError

    def delete_model(self, team_id: int, model_id: int):
        raise NotImplementedError

    def get_ml_model_download_tokens(
        self, team_id: int, model_id: int
    ) -> ServiceResponse:
        raise NotImplementedError

    def run_prediction(
        self, team_id: int, project_id: int, ml_model_id: int, image_ids: list
    ):
        raise NotImplementedError

    def delete_image_annotations(
        self,
        team_id: int,
        project_id: int,
        folder_id: int = None,
        image_names: List[str] = None,
    ) -> dict:
        raise NotImplementedError

    def get_annotations_delete_progress(
        self, team_id: int, project_id: int, poll_id: int
    ):
        raise NotImplementedError

    def get_limitations(
        self, team_id: int, project_id: int, folder_id: int = None
    ) -> ServiceResponse:
        raise NotImplementedError

    @abstractmethod
    def get_annotations(
        self,
        project_id: int,
        team_id: int,
        folder_id: int,
        items: List[str],
        reporter: Reporter,
        callback: Callable = None,
    ) -> List[dict]:
        raise NotImplementedError

    @abstractmethod
    async def download_annotations(
        self,
        project_id: int,
        team_id: int,
        folder_id: int,
        reporter: Reporter,
        download_path: str,
        postfix: str,
        items: List[str] = None,
        callback: Callable = None,
    ) -> int:
        """
        Returns the number of items downloaded
        """
        raise NotImplementedError

    def upload_priority_scores(
        self, team_id: int, project_id: int, folder_id: int, priorities: list
    ) -> dict:
        raise NotImplementedError

    def get_integrations(self, team_id: int) -> List[dict]:
        raise NotImplementedError

    def attach_integrations(
        self,
        team_id: int,
        project_id: int,
        integration_id: int,
        folder_id: int,
        folder_name: str,
    ) -> bool:
        raise NotImplementedError

    def saqul_query(
        self,
        team_id: int,
        project_id: int,
        folder_id: int,
        query: str = None,
        subset_id: int = None,
    ) -> ServiceResponse:
        raise NotImplementedError

    def validate_saqul_query(self, team_id: int, project_id: int, query: str) -> dict:
        raise NotImplementedError

    def list_sub_sets(self, team_id: int, project_id: int) -> ServiceResponse:
        raise NotImplementedError

    def create_custom_schema(
        self, team_id: int, project_id: int, schema: dict
    ) -> ServiceResponse:
        raise NotImplementedError

    def get_custom_schema(self, team_id: int, project_id: int) -> ServiceResponse:
        raise NotImplementedError

    def delete_custom_schema(
        self, team_id: int, project_id: int, fields: List[str]
    ) -> ServiceResponse:
        raise NotImplementedError

    def upload_custom_fields(
        self, team_id: int, project_id: int, folder_id: int, items: List[dict]
    ) -> ServiceResponse:
        raise NotImplementedError

    def delete_custom_fields(
        self,
        team_id: int,
        project_id: int,
        folder_id: int,
        items: List[Dict[str, List[str]]],
    ) -> ServiceResponse:
        raise NotImplementedError



import json
import os
from typing import Callable

import aiohttp
from lib.core.reporter import Reporter


class StreamedAnnotations:
    DELIMITER = b"\\n;)\\n"

    def __init__(
        self,
        headers: dict,
        reporter: Reporter,
        callback: Callable = None,
        map_function: Callable = None,
    ):
        self._headers = headers
        self._annotations = []
        self._reporter = reporter
        self._callback: Callable = callback
        self._map_function = map_function

    async def fetch(
        self,
        method: str,
        session: aiohttp.ClientSession,
        url: str,
        data: dict = None,
        params: dict = None,
    ):
        kwargs = {"params": params}
        if data:
            kwargs["json"] = data
        response = await session._request(method, url, **kwargs)
        buffer = b""
        async for line in response.content.iter_any():
            slices = line.split(self.DELIMITER)
            if len(slices) == 1:
                buffer += slices[0]
                continue
            elif slices[0]:
                self._reporter.update_progress()
                yield json.loads(buffer + slices[0])
            for data in slices[1:-1]:
                self._reporter.update_progress()
                yield json.loads(data)
            buffer = slices[-1]
        if buffer:
            yield json.loads(buffer)
            self._reporter.update_progress()

    async def get_data(
        self,
        url: str,
        data: list,
        method: str = "post",
        params=None,
        chunk_size: int = 100,
        verify_ssl: bool = False,
    ):
        async with aiohttp.ClientSession(
            headers=self._headers,
            connector=aiohttp.TCPConnector(ssl=verify_ssl),
        ) as session:
            if chunk_size:
                for i in range(0, len(data), chunk_size):
                    data_to_process = data[i : i + chunk_size]
                    async for annotation in self.fetch(
                        method,
                        session,
                        url,
                        self._process_data(data_to_process),
                        params=params,
                    ):
                        self._annotations.append(
                            self._callback(annotation) if self._callback else annotation
                        )
            else:
                async for annotation in self.fetch(
                    method, session, url, self._process_data(data), params=params
                ):
                    self._annotations.append(
                        self._callback(annotation) if self._callback else annotation
                    )
        return self._annotations

    @staticmethod
    def _store_annotation(path, postfix, annotation: dict, callback: Callable = None):
        os.makedirs(path, exist_ok=True)
        with open(f"{path}/{annotation['metadata']['name']}{postfix}", "w") as file:
            annotation = callback(annotation) if callback else annotation
            json.dump(annotation, file)

    def _process_data(self, data):
        if data and self._map_function:
            return self._map_function(data)
        return data

    async def download_data(
        self,
        url: str,
        data: list,
        download_path: str,
        postfix: str,
        session,
        method: str = "post",
        params=None,
        chunk_size: int = 5000,
    ) -> int:
        """
        Returns the number of items downloaded
        """
        items_downloaded: int = 0
        if chunk_size and data:
            for i in range(0, len(data), chunk_size):
                data_to_process = data[i : i + chunk_size]
                async for annotation in self.fetch(
                    method,
                    session,
                    url,
                    self._process_data(data_to_process),
                    params=params,
                ):
                    self._store_annotation(
                        download_path,
                        postfix,
                        annotation,
                        self._callback,
                    )
                    items_downloaded += 1
        else:
            async for annotation in self.fetch(
                method, session, url, self._process_data(data), params=params
            ):
                self._store_annotation(
                    download_path, postfix, annotation, self._callback
                )
                items_downloaded += 1
        return items_downloaded



from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from pydantic import BaseModel
from pydantic import Extra
from pydantic import parse_obj_as


class Limit(BaseModel):
    max_image_count: Optional[int]
    remaining_image_count: int


class UserLimits(BaseModel):
    user_limit: Optional[Limit]
    project_limit: Limit
    folder_limit: Limit


class UploadAnnotationAuthData(BaseModel):
    access_key: str
    secret_key: str
    session_token: str
    region: str
    bucket: str
    images: Dict[int, dict]

    class Config:
        extra = Extra.allow
        fields = {
            "access_key": "accessKeyId",
            "secret_key": "secretAccessKey",
            "session_token": "sessionToken",
            "region": "region",
        }

    def __init__(self, **data):
        credentials = data["creds"]
        data.update(credentials)
        del data["creds"]
        super().__init__(**data)


class DownloadMLModelAuthData(BaseModel):
    access_key: str
    secret_key: str
    session_token: str
    region: str
    bucket: str
    paths: List[str]

    class Config:
        extra = Extra.allow
        fields = {
            "access_key": "accessKeyId",
            "secret_key": "secretAccessKey",
            "session_token": "sessionToken",
            "region": "region",
        }

    def __init__(self, **data):
        credentials = data["tokens"]
        data.update(credentials)
        del data["tokens"]
        super().__init__(**data)


class UploadCustomFieldValues(BaseModel):
    succeeded_items: Optional[List[Any]]
    failed_items: Optional[List[str]]
    error: Optional[Any]


class ServiceResponse(BaseModel):
    status: int
    reason: str
    content: Union[bytes, str]
    data: Any
    count: Optional[int] = 0
    _error: str

    class Config:
        extra = Extra.allow

    def __init__(self, response, content_type=None, dispatcher: Callable = None):
        data = {
            "status": response.status_code,
            "reason": response.reason,
            "content": response.content,
        }
        try:
            response_json = response.json()
        except Exception:
            response_json = dict()
        if not response.ok:
            data["_error"] = response_json.get("error", "Unknown Error")
            data["data"] = response_json
            super().__init__(**data)
            return
        if dispatcher:
            _data = response_json
            response_json = dispatcher(_data)
            data.update(_data)
        try:
            if content_type and content_type is not self.__class__:
                data["data"] = parse_obj_as(content_type, response_json)
            else:
                data["data"] = response_json
        except Exception:
            data["data"] = {}

        super().__init__(**data)

    @property
    def status_code(self):
        return self.status

    @property
    def ok(self):
        return 199 < self.status < 300

    @property
    def error(self):
        if self._error:
            return self._error
        default_message = self.reason if self.reason else "Unknown Error"
        if isinstance(self.data, dict) and "error" in self.data:
            return self.data.get("error", default_message)
        else:
            return getattr(self.data, "error", default_message)



import os
import sys

__version__ = "4.4.1"

sys.path.append(os.path.split(os.path.realpath(__file__))[0])

import logging.config  # noqa
import requests  # noqa
from packaging.version import parse  # noqa
from superannotate.lib.app.input_converters import convert_json_version  # noqa
from superannotate.lib.app.input_converters import convert_project_type  # noqa
from superannotate.lib.app.analytics.class_analytics import class_distribution  # noqa
from superannotate.lib.app.exceptions import AppException  # noqa
from superannotate.lib.app.input_converters import convert_json_version  # noqa
from superannotate.lib.app.input_converters import convert_project_type  # noqa
from superannotate.lib.app.input_converters import export_annotation  # noqa
from superannotate.lib.app.input_converters import import_annotation  # noqa
from superannotate.lib.app.interface.sdk_interface import SAClient  # noqa
from superannotate.lib.core import PACKAGE_VERSION_INFO_MESSAGE  # noqa
from superannotate.lib.core import PACKAGE_VERSION_MAJOR_UPGRADE  # noqa
from superannotate.lib.core import PACKAGE_VERSION_UPGRADE  # noqa
from superannotate.logger import get_default_logger  # noqa
import superannotate.lib.core.enums as enums  # noqa

SESSIONS = {}

__all__ = [
    "__version__",
    "SAClient",
    # Utils
    "enums",
    "AppException",
    # analytics
    "class_distribution",
    # converters
    "convert_json_version",
    "import_annotation",
    "export_annotation",
    "convert_project_type",
]

__author__ = "Superannotate"

logging.getLogger("botocore").setLevel(logging.CRITICAL)
logger = get_default_logger()


def log_version_info():
    local_version = parse(__version__)
    if local_version.is_prerelease:
        logger.info(PACKAGE_VERSION_INFO_MESSAGE.format(__version__))
    req = requests.get("https://pypi.python.org/pypi/superannotate/json")
    if req.ok:
        releases = req.json().get("releases", [])
        pip_version = parse("0")
        for release in releases:
            ver = parse(release)
            if not ver.is_prerelease or local_version.is_prerelease:
                pip_version = max(pip_version, ver)
        if pip_version.major > local_version.major:
            logger.warning(
                PACKAGE_VERSION_MAJOR_UPGRADE.format(local_version, pip_version)
            )
        elif pip_version > local_version:
            logger.warning(PACKAGE_VERSION_UPGRADE.format(local_version, pip_version))


if not os.environ.get("SA_VERSION_CHECK", "True").lower() == "false":
    log_version_info()



import functools
import os
import sys
from inspect import signature
from pathlib import Path
from types import FunctionType
from typing import Iterable
from typing import Sized
from typing import Tuple

import lib.core as constants
from lib.app.helpers import extract_project_folder
from lib.app.interface.types import validate_arguments
from lib.core import CONFIG
from lib.core.exceptions import AppException
from lib.infrastructure.controller import Controller
from lib.infrastructure.repositories import ConfigRepository
from mixpanel import Mixpanel
from superannotate import __version__


class BaseInterfaceFacade:
    REGISTRY = []

    def __init__(self, token: str = None, config_path: str = None):
        version = os.environ.get("SA_VERSION", "v1")
        _token, _config_path = None, None
        _host = os.environ.get("SA_URL", constants.BACKEND_URL)
        _ssl_verify = not os.environ.get("SA_SSL", "True").lower() in (
            "false",
            "f",
            "0",
        )
        if token:
            _token = Controller.validate_token(token=token)
        elif config_path:
            _token, _host, _ssl_verify = self._retrieve_configs(config_path)
        else:
            _token = os.environ.get("SA_TOKEN")
            if not _token:
                _token, _host, _ssl_verify = self._retrieve_configs(
                    constants.CONFIG_PATH
                )
        _host = _host if _host else constants.BACKEND_URL
        _ssl_verify = True if _ssl_verify is None else _ssl_verify

        self._token, self._host = _token, _host
        self.controller = Controller(_token, _host, _ssl_verify, version)
        BaseInterfaceFacade.REGISTRY.append(self)

    @staticmethod
    def _retrieve_configs(path) -> Tuple[str, str, str]:
        config_path = os.path.expanduser(str(path))
        if not Path(config_path).is_file() or not os.access(config_path, os.R_OK):
            raise AppException(
                f"SuperAnnotate config file {str(config_path)} not found."
            )
        config_repo = ConfigRepository(config_path)
        return (
            Controller.validate_token(config_repo.get_one("token").value),
            config_repo.get_one("main_endpoint").value,
            config_repo.get_one("ssl_verify").value,
        )

    @property
    def host(self):
        return self._host

    @property
    def token(self):
        return self._token


class Tracker:
    def get_mp_instance(self) -> Mixpanel:
        client = self.get_client()
        mp_token = "ca95ed96f80e8ec3be791e2d3097cf51"
        if client:
            if client.host != constants.BACKEND_URL:
                mp_token = "e741d4863e7e05b1a45833d01865ef0d"
        return Mixpanel(mp_token)

    @staticmethod
    def get_default_payload(team_name, user_id):
        return {
            "SDK": True,
            "Team": team_name,
            "Team Owner": user_id,
            "Version": __version__,
        }

    def __init__(self, function):
        self.function = function
        self._client = None
        functools.update_wrapper(self, function)

    def get_client(self):
        if not self._client:
            if BaseInterfaceFacade.REGISTRY:
                return BaseInterfaceFacade.REGISTRY[-1]
            else:
                from lib.app.interface.sdk_interface import SAClient

                try:
                    return SAClient()
                except Exception:
                    pass
        elif hasattr(self._client, "controller"):
            return self._client

    @staticmethod
    def extract_arguments(function, *args, **kwargs) -> dict:
        bound_arguments = signature(function).bind(*args, **kwargs)
        bound_arguments.apply_defaults()
        return dict(bound_arguments.arguments)

    @staticmethod
    def default_parser(function_name: str, kwargs: dict) -> tuple:
        properties = {}
        for key, value in kwargs.items():
            if key == "self":
                continue
            elif value is None:
                properties[key] = value
            elif key == "project":
                properties["project_name"], folder_name = extract_project_folder(value)
                if folder_name:
                    properties["folder_name"] = folder_name
            elif isinstance(value, (str, int, float, bool)):
                properties[key] = value
            elif isinstance(value, dict):
                properties[key] = list(value.keys())
            elif isinstance(value, Sized):
                properties[key] = len(value)
            elif isinstance(value, Iterable):
                properties[key] = "N/A"
            else:
                properties[key] = str(value)
        return function_name, properties

    def _track(self, user_id: str, event_name: str, data: dict):
        if "pytest" not in sys.modules:
            self.get_mp_instance().track(user_id, event_name, data)

    def _track_method(self, args, kwargs, success: bool):
        try:
            client = self.get_client()
            if not client:
                return
            function_name = self.function.__name__ if self.function else ""
            arguments = self.extract_arguments(self.function, *args, **kwargs)
            event_name, properties = self.default_parser(function_name, arguments)
            user_id = client.controller.team_data.creator_id
            team_name = client.controller.team_data.name

            properties["Success"] = success
            default = self.get_default_payload(team_name=team_name, user_id=user_id)
            self._track(
                user_id,
                event_name,
                {**default, **properties, **CONFIG.get_current_session().data},
            )
        except BaseException:
            pass

    def __get__(self, obj, owner=None):
        if obj is not None:
            self._client = obj
            tmp = functools.partial(self.__call__, obj)
            functools.update_wrapper(tmp, self.function)
            return tmp
        return self

    def __call__(self, *args, **kwargs):
        success = True
        try:
            result = self.function(*args, **kwargs)
        except Exception as e:
            success = False
            raise e
        else:
            return result
        finally:
            self._track_method(args=args, kwargs=kwargs, success=success)


class TrackableMeta(type):
    def __new__(mcs, name, bases, attrs):
        for attr_name, attr_value in attrs.items():
            if isinstance(
                attr_value, FunctionType
            ) and not attr_value.__name__.startswith("_"):
                attrs[attr_name] = Tracker(validate_arguments(attr_value))
        tmp = super().__new__(mcs, name, bases, attrs)
        return tmp



import itertools
import math
from collections import defaultdict
from typing import Any
from typing import List
from typing import Optional

from lib.core.enums import AnnotationTypes
from pydantic import BaseModel


class Annotation(BaseModel):
    instanceId: int
    type: str
    className: Optional[str]
    classId: Optional[int]
    x: Optional[Any]
    y: Optional[Any]
    points: Any
    attributes: Optional[List[Any]] = []
    keyframe: bool = False


class FrameAnnotation(BaseModel):
    frame: int
    annotations: List[Annotation] = []


class VideoFrameGenerator:
    def __init__(self, annotation_data: dict, fps: int):
        self.id_generator = iter(itertools.count(0))
        self._annotation_data = annotation_data
        duration = annotation_data["metadata"]["duration"]
        duration = 0 if not duration else duration
        self.duration = duration / (1000 * 1000)
        self.fps = fps
        self.ratio = 1000 * 1000 / fps
        self._frame_id = 1
        self.frames_count = int(math.ceil(self.duration * fps))
        self.annotations: dict = {}
        self._mapping = {}
        self._process()

    def get_frame(self, frame_no: int):
        try:
            return self.annotations[frame_no]
        except KeyError:
            self.annotations[frame_no] = FrameAnnotation(frame=frame_no)
            return self.annotations[frame_no]

    def _interpolate(
        self,
        class_name: str,
        class_id: int,
        from_frame: int,
        to_frame: int,
        data: dict,
        instance_id: int,
        steps: dict = None,
        annotation_type: str = "bbox",
    ) -> dict:
        annotations = {}
        for idx, frame_idx in enumerate(range(from_frame + 1, to_frame), 1):
            tmp_data = {}
            if annotation_type == AnnotationTypes.BBOX and data.get("points") and steps:
                tmp_data["points"] = {
                    "x1": round(data["points"]["x1"] + steps["x1"] * idx, 2),
                    "y1": round(data["points"]["y1"] + steps["y1"] * idx, 2),
                    "x2": round(data["points"]["x2"] + steps["x2"] * idx, 2),
                    "y2": round(data["points"]["y2"] + steps["y2"] * idx, 2),
                }
            elif annotation_type == AnnotationTypes.POINT:
                tmp_data = {
                    "x": round(data["x"] + steps["x"] * idx, 2),
                    "y": round(data["y"] + steps["y"] * idx, 2),
                }
            elif annotation_type != AnnotationTypes.EVENT:
                tmp_data["points"] = data["points"]

            annotations[frame_idx] = Annotation(
                instanceId=instance_id,
                type=annotation_type,
                className=class_name,
                classId=class_id,
                attributes=data["attributes"],
                keyframe=False,
                **tmp_data
            )
        return annotations

    def _add_annotation(self, frame_no: int, annotation: Annotation):

        frame = self.get_frame(frame_no)
        frame.annotations.append(annotation)

    @staticmethod
    def pairwise(data: list):
        a, b = itertools.tee(data)
        next(b, None)
        return zip(a, b)

    def get_median(self, annotations: List[dict]) -> dict:
        if len(annotations) >= 1:
            return annotations[0]
        # Let's just leave the code for reference.
        # first_annotations = annotations[:1][0]
        # median = (
        #     first_annotations["timestamp"] // self.ratio
        # ) * self.ratio + self.ratio / 2
        # median_annotation = first_annotations
        # distance = abs(median - first_annotations["timestamp"])
        # for annotation in annotations[1:]:
        #     annotation_distance = abs(median - annotation["timestamp"])
        #     if annotation_distance < distance:
        #         distance = annotation_distance
        #         median_annotation = annotation
        # return median_annotation

    @staticmethod
    def merge_first_frame(frames_mapping):
        try:
            if 0 in frames_mapping:
                frames_mapping[0].extend(frames_mapping[1])
                frames_mapping[1] = frames_mapping[0]
                del frames_mapping[0]
        finally:
            return frames_mapping

    def _interpolate_frames(
        self,
        from_frame,
        from_frame_no,
        to_frame,
        to_frame_no,
        annotation_type,
        class_name,
        class_id,
        instance_id,
    ):
        steps = None
        frames_diff = to_frame_no - from_frame_no
        if (
            annotation_type == AnnotationTypes.BBOX
            and from_frame.get("points")
            and to_frame.get("points")
        ):
            steps = {}
            for point in "x1", "x2", "y1", "y2":
                steps[point] = round(
                    (to_frame["points"][point] - from_frame["points"][point])
                    / frames_diff,
                    2,
                )
        elif annotation_type == AnnotationTypes.POINT:
            steps = {
                "x": (to_frame["x"] - from_frame["x"]) / frames_diff,
                "y": (to_frame["y"] - from_frame["y"]) / frames_diff,
            }
        elif annotation_type in (AnnotationTypes.POLYGON, AnnotationTypes.POLYLINE):
            steps = [
                (to_point - from_point) / frames_diff
                for from_point, to_point in zip(
                    from_frame["points"], to_frame["points"]
                )
            ]
        return self._interpolate(
            class_name=class_name,
            class_id=class_id,
            from_frame=from_frame_no,
            to_frame=to_frame_no,
            data=from_frame,
            instance_id=instance_id,
            steps=steps,
            annotation_type=annotation_type,
        )

    def _process(self):
        for instance in self._annotation_data["instances"]:
            instance_id = next(self.id_generator)
            annotation_type = instance["meta"]["type"]
            class_name = instance["meta"].get("className")
            class_id = instance["meta"].get("classId", -1)
            for parameter in instance["parameters"]:
                frames_mapping = defaultdict(list)
                interpolated_frames = {}
                for timestamp in parameter["timestamps"]:
                    frames_mapping[
                        int(math.ceil(timestamp["timestamp"] / self.ratio))
                    ].append(timestamp)
                frames_mapping = self.merge_first_frame(frames_mapping)
                for from_frame_no, to_frame_no in self.pairwise(sorted(frames_mapping)):
                    last_frame_no = to_frame_no
                    from_frame, to_frame = (
                        frames_mapping[from_frame_no][-1],
                        frames_mapping[to_frame_no][0],
                    )
                    frames_diff = to_frame_no - from_frame_no
                    if frames_diff > 1:
                        interpolated_frames.update(
                            self._interpolate_frames(
                                from_frame=from_frame,
                                from_frame_no=from_frame_no,
                                to_frame=to_frame,
                                to_frame_no=to_frame_no,
                                class_name=class_name,
                                class_id=class_id,
                                annotation_type=annotation_type,
                                instance_id=instance_id,
                            )
                        )

                    start_median_frame = self.get_median(frames_mapping[from_frame_no])
                    end_median_frame = self.get_median(frames_mapping[to_frame_no])
                    for frame_no, frame in (
                        (from_frame_no, start_median_frame),
                        (last_frame_no, end_median_frame),
                    ):
                        interpolated_frames[frame_no] = Annotation(
                            instanceId=instance_id,
                            type=annotation_type,
                            className=class_name,
                            classId=class_id,
                            x=frame.get("x"),
                            y=frame.get("y"),
                            points=frame.get("points"),
                            attributes=frame["attributes"],
                            keyframe=True,
                        )
                if frames_mapping and not interpolated_frames:
                    median = self.get_median(frames_mapping[1])
                    interpolated_frames[1] = Annotation(
                        instanceId=instance_id,
                        type=annotation_type,
                        className=class_name,
                        classId=class_id,
                        x=median.get("x"),
                        y=median.get("y"),
                        points=median.get("points"),
                        attributes=median["attributes"],
                        keyframe=True,
                    )

                for frame_no, annotation in interpolated_frames.items():
                    self._add_annotation(frame_no, annotation)

    def __iter__(self):
        for frame_no in range(1, int(self.frames_count) + 1):
            frame = self.get_frame(frame_no)
            yield {**frame.dict(exclude_unset=True), **frame.dict(exclude_none=True)}



import os
import re

from setuptools import find_packages, setup


def get_version():
    init = open(os.path.join(os.path.dirname(__file__), 'src', 'superannotate', '__init__.py')).read()
    version_re = re.compile(r'''__version__ = ["']((\d+)\.(\d+)\.(\d+)((dev(\d+))?(b(\d+))?))['"]''')
    return version_re.search(init).group(1)


sdk_version = get_version()


requirements_path = "requirements_{}.txt".format('dev' if 'dev' in sdk_version else 'prod')

requirements = []

with open("requirements.txt") as f:
    requirements.extend(f.read().splitlines())

with open(requirements_path) as f:
    requirements.extend(f.read().splitlines())


with open('README.md') as f:
    readme = f.read()

readme = "\n".join(readme.split('\n')[2:])


setup(
    name='superannotate',
    version=sdk_version,
    package_dir={"": "src"},
    package_data={"superannotate": ["logging.conf"]},
    packages=find_packages(where="src"),
    description='Python SDK to SuperAnnotate platform',
    license='MIT',
    author='SuperAnnotate AI',
    url='https://github.com/superannotateai/superannotate-python-sdk',
    long_description=readme,
    long_description_content_type='text/markdown',
    install_requires=requirements,
    setup_requires=['wheel'],
    description_file="README.md",
    entry_points={
        'console_scripts': ['superannotatecli = superannotate.lib.app.bin.superannotate:main']
    },
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    project_urls={
        'Documentation': 'https://superannotate.readthedocs.io/en/stable/',
    },
    python_requires='>=3.7'
)



from typing import Dict
from typing import List

from lib.core.entities import FolderEntity
from lib.core.entities import ProjectEntity
from lib.core.reporter import Reporter
from lib.core.response import Response
from lib.core.serviceproviders import SuperannotateServiceProvider
from lib.core.usecases import BaseReportableUseCase


class CreateCustomSchemaUseCase(BaseReportableUseCase):
    def __init__(
        self,
        reporter: Reporter,
        project: ProjectEntity,
        schema: dict,
        backend_client: SuperannotateServiceProvider,
    ):
        super().__init__(reporter)
        self._project = project
        self._schema = schema
        self._backend_client = backend_client

    def execute(self) -> Response:
        response = self._backend_client.create_custom_schema(
            team_id=self._project.team_id,
            project_id=self._project.id,
            schema=self._schema,
        )
        if response.ok:
            self._response.data = response.data
        else:
            errors = response.data.get("errors")
            if errors:
                separator = "\n- "
                report = separator + separator.join(errors)
            else:
                report = response.error
            self._response.errors = report
        return self._response


class GetCustomSchemaUseCase(BaseReportableUseCase):
    def __init__(
        self,
        reporter: Reporter,
        project: ProjectEntity,
        backend_client: SuperannotateServiceProvider,
    ):
        super().__init__(reporter)
        self._project = project
        self._backend_client = backend_client

    def execute(self) -> Response:
        response = self._backend_client.get_custom_schema(
            team_id=self._project.team_id,
            project_id=self._project.id,
        )
        if response.ok:
            self._response.data = response.data
        else:
            self._response.errors = response.error
        return self._response


class DeleteCustomSchemaUseCase(BaseReportableUseCase):
    def __init__(
        self,
        reporter: Reporter,
        project: ProjectEntity,
        fields: List[str],
        backend_client: SuperannotateServiceProvider,
    ):
        super().__init__(reporter)
        self._project = project
        self._fields = fields
        self._backend_client = backend_client

    def execute(self) -> Response:
        if self._fields:
            self.reporter.log_info("Matched fields deleted from schema.")
        response = self._backend_client.delete_custom_schema(
            team_id=self._project.team_id,
            project_id=self._project.id,
            fields=self._fields,
        )
        if response.ok:
            use_case_response = GetCustomSchemaUseCase(
                reporter=self.reporter,
                project=self._project,
                backend_client=self._backend_client,
            ).execute()
            if use_case_response.errors:
                self._response.errors = use_case_response.errors
            else:
                self._response.data = use_case_response.data
        else:
            self._response.errors = response.error
        return self._response


class UploadCustomValuesUseCase(BaseReportableUseCase):
    CHUNK_SIZE = 5000

    def __init__(
        self,
        reporter: Reporter,
        project: ProjectEntity,
        folder: FolderEntity,
        items: List[Dict[str, str]],
        backend_client: SuperannotateServiceProvider,
    ):
        super().__init__(reporter)
        self._project = project
        self._folder = folder
        self._items = items
        self._backend_client = backend_client

    def execute(self) -> Response:
        uploaded_items, failed_items = [], []
        self.reporter.log_info(
            "Validating metadata against the schema of the custom fields. "
            "Valid metadata will be attached to the specified item."
        )
        with self.reporter.spinner:
            for idx in range(0, len(self._items), self.CHUNK_SIZE):
                response = self._backend_client.upload_custom_fields(
                    project_id=self._project.id,
                    team_id=self._project.team_id,
                    folder_id=self._folder.uuid,
                    items=self._items[idx : idx + self.CHUNK_SIZE],  # noqa: E203
                )
                if not response.ok:
                    self._response.errors = response.error
                    return self._response
                failed_items.extend(response.data.failed_items)

        if failed_items:
            self.reporter.log_error(
                f"The metadata dicts of {len(failed_items)} items are invalid because they don't match "
                f'the schema of the custom fields defined for the "{self._project.name}" project.'
            )
        self._response.data = {
            "succeeded": list(
                {list(item)[0] for item in self._items} ^ set(failed_items)
            ),
            "failed": failed_items,
        }
        return self._response


class DeleteCustomValuesUseCase(BaseReportableUseCase):
    CHUNK_SIZE = 5000

    def __init__(
        self,
        reporter: Reporter,
        project: ProjectEntity,
        folder: FolderEntity,
        items: List[Dict[str, List[str]]],
        backend_client: SuperannotateServiceProvider,
    ):
        super().__init__(reporter)
        self._project = project
        self._folder = folder
        self._items = items
        self._backend_client = backend_client

    def execute(self) -> Response:
        for idx in range(0, len(self._items), self.CHUNK_SIZE):
            response = self._backend_client.delete_custom_fields(
                project_id=self._project.id,
                team_id=self._project.team_id,
                folder_id=self._folder.uuid,
                items=self._items[idx : idx + self.CHUNK_SIZE],  # noqa: E203
            )
            if not response.ok:
                self._response.errors = response.error
                return self._response
        self.reporter.log_info(
            "Corresponding fields and their values removed from items."
        )
        return self._response
